"""
Supplementary File: Script for Experiment 4.

This Python 3 script extends Futrell et al.'s (2020) noisy surprisal model:
    Futrell, Richard, Edward Gibson & Roger Levy. 2020. Lossy-context surprisal: An information-theoretic model of memory effects in sentence processing. Cognitive Science 44(3). 1-54.
    Model: https://github.com/Futrell/lc-surprisal

This script reuses functions from Futrell et al.'s "experiments.py". It calculates surprisal values using the noisy surprisal model of English, German/Dutch, and Mandarin Chinese.

It adds three functions for modeling Mandarin Chinese:
    embedding_lang_MC: the Mandarin toy grammar, cf. embedding_lang
    verb_forgetting_conditions_MC: calculate surprisal values for Mandarin prefixes, cf. verb_forgetting_conditions
    summary: print formatted surprisal values

Instructions:
1.  Install/download the Python files from Futrell et al.'s Github link (above).
2.  Place this file in the same folder.
3.  Run Python from that folder. Enter the following:
    
        import exp4script as exp4 # more generally, "import [name of this file] as exp4"
        # Parameters:
        # m = rate of NP --> N XP, where XP is a relative clause or some other material
        # r = rate of NP --> N RC relative to NP --> N XP, where XP is a relative clause or some other material
        # ee = deletion rate (default = 0.2)
        # ss = rate a noun appears with a V-NP relative clause (so ss = 0 for German, ss = 0.8 for English and Mandarin)
        # Note: m and r should remain unchanged.
        # In Futrell et al. 2020, ee is called "d" and ss is called "1-f"
        exp4.summary(m=.5, r=.5, ee=0.2, ss=0.8) # This generates Table 8
"""
from experiments import * # Import all functions, parameter values from Futrell et al.'s experiments.py file

# Toy grammar for Mandarin
# p_mod = m; p_rc = r; p_src = ss
def embedding_lang_MC(p_mod, p_rc, p_src, num_levels):
    rules = [
        (pcfg.Rule('S', ('N', 'V', 'NP')), 1), # Assume all sentences are transitive sentences
        (pcfg.Rule('NP', ('N',)), 1 - p_mod),
        (pcfg.Rule('NP', ('RC', 'N')), p_mod * p_rc), # RCs precede nouns in MC
        (pcfg.Rule('NP', ('N', 'ConjP')), p_mod * (1 - p_rc)), # Assume non-RC modification to be coordination
        (pcfg.Rule('ConjP', ('Conj', 'NP')), 1),
        (pcfg.Rule('RC', ('NP', 'V', 'C')), 1 - p_src), # Object RCs in MC: Subject-verb-DE word order
        (pcfg.Rule('RC', ('V', 'NP', 'C')), p_src) # Subject RCs in MC: Verb-object-DE word order
    ]
    grammar = pcfg.make_bounded_pcfg(SymbolicEnumeration, rules, num_levels)
    return grammar.distribution()

# Verb forgetting, with Mandarin input strings. This is different from the original function in 3 places.
def verb_forgetting_conditions_MC(num_levels=2,
                               m=sympy.Symbol('m'),
                               r=sympy.Symbol('r'),
                               s=sympy.Symbol('s'),
                               e=sympy.Symbol('e')):

    SE = SymbolicEnumeration

    # create the language ## Difference #1: load the MC toy grammar
    lang = embedding_lang_MC(m, r, s, num_levels).marginalize()

    # the noise is sequence erasure noise, here in the SE monad
    def symbol_noise(x):
        # flip e, if true then erase else don't
        return SE.flip(e).bind( 
            SE.lift_ret(lambda b: 'E' if b else x)
        )

    def erasure_noise(s):
        return SE.mapM(symbol_noise, s)

    def deletion_noise(s):
        return SE.mapM(symbol_noise, s).bind(
            SE.lift_ret(lambda s: tuple(x for x in s if x != 'E'))
        )

    # apply the noise function to each sentence.
    # this might be slow.
    npt = incnoise.noisy_prefix_tree(lang, deletion_noise)

    prefixes = [ ## Difference #2: Generate prefixes consistent with Chinese word order: NVVN, NVVVNCN
        tuple("NV")
        + tuple("V") * (num_levels - k)
        + tuple("NC") * (num_levels - k - 1) + tuple("N")
        for k in reversed(range(num_levels))]

    def gen():
        for i in range(num_levels):
            noisy_prefix_distro = deletion_noise(prefixes[i]).marginalize()

            cost_correct = 0
            cost_incorrect = 0
            for noisy_prefix, p in noisy_prefix_distro.dict.items():
                p_correct = npt[noisy_prefix].dict.get('C', 0) ## Difference #3: grammatical continuation involves C ('de')
                p_incorrect = npt[noisy_prefix].dict.get(incnoise.HALT, 0)
                cost_correct -= p * sympy.log(p_correct)
                cost_incorrect -= p * sympy.log(p_incorrect)
            cost_correct = cost_correct.simplify()
            cost_incorrect = cost_incorrect.simplify()
            yield cost_correct, cost_incorrect

    return tuple(gen())

# Print out a formatted summary
def summary(m=.5, r=.5, ss = 0.8, ee = 0.2):
    print('Calculating ...')
    _, english = verb_forgetting_conditions(m=.5, r=.5, e=ee, s=ss)
    print('English')
    _, german = verb_forgetting_conditions(m=.5, r=.5, e=ee, s=0)
    print('German/Dutch')
    _, mc = verb_forgetting_conditions_MC(m=.5, r=.5, e=ee, s=ss)
    print('Mandarin')

    print("\nP(NP --> N XP) or P(NP --> XP N) for MC:\t", m, 
          "\nP(XP is a RC):\t", r, 
          "\nPercentage RCs with V-NP word order for EN and MC:\t", ss, 
          "\nDeletion rate:\t", ee, '\n')
    print('Language\t', 'Grm.\t', 'Ungrm.\t', 'Grm.-Ungrm.\t')
    for lg, surprisal in {'English': english, 'German/Dutch': german, 'Mandarin': mc}.items():
        print(lg, '\t', 
              # Report values in bits
              round(surprisal[0]/log(2), 3), '\t', 
              round(surprisal[1]/log(2), 3), '\t', 
              round(surprisal[0]/log(2) - surprisal[1]/log(2), 3))
        
if __name__ == '__main__':
    import nose
    nose.runmodule()
